﻿using System;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace simplehttpserver
{
    /// <summary>
    /// author: anbo<br />
    /// References
    /// <list type="">
    /// <item>http://msdn.microsoft.com/en-us/library/b6xa24z5(v=vs.110).aspx</item>
    /// <item>http://www.codeproject.com/Articles/17071/Sample-HTTP-Server-Skeleton-in-C</item>
    /// <item>http://msdn.microsoft.com/en-us/library/aa989072(v=vs.80).aspx</item>
    /// <item>http://stackoverflow.com/questions/12009695/net-equivalent-of-javas-bufferedreader</item>
    /// </list>
    /// </summary>
    public class HttpServer
    {
        /// <summary>
        /// The default port of the HTTP server
        /// </summary>
        public static readonly int DefaultPort = 8888;
        private static readonly string RootCatalog = "c:/temp";
        private readonly int _serverPort;
        private volatile Boolean _keepOnRunning = true;

        /// <summary>
        /// Constructs and starts an instance of the HTTP server
        /// </summary>
        /// <param name="serverPort">port number of the HTTP server</param>
        public HttpServer(int serverPort)
        {
            if (serverPort < 0 || serverPort > 65535)
            {
                throw new ArgumentOutOfRangeException("serverPort", serverPort, "port number out of range");
            }
            this._serverPort = serverPort;
        }

        public void RunServer()
        {
            IPHostEntry ipHostEntry = Dns.GetHostEntry("localhost");
            IPAddress localIpAddress = ipHostEntry.AddressList[0];
            TcpListener tcpListener = new TcpListener(localIpAddress, _serverPort);
            tcpListener.Start();
            Console.WriteLine("Server listening on port: " + _serverPort);
            while (_keepOnRunning)
            {
                try
                {
                    while (!tcpListener.Pending())
                    {
                        Thread.Sleep(100);
                    }
                    TcpClient client = tcpListener.AcceptTcpClient();
                    Task.Run(() => HandleRequest(client));
                }
                catch (SocketException ex)
                {
                    Console.WriteLine("Serious problem: " + ex.Message);
                }
            }
            tcpListener.Stop();
        }

        public void Stop()
        {
            _keepOnRunning = false;
        }

        private static void HandleRequest(TcpClient tcpClient)
        {
            try
            {
                NetworkStream clientStream = tcpClient.GetStream();
                using (StreamReader fromClient = new StreamReader(clientStream, Encoding.UTF8))
                using (StreamWriter toClient = new StreamWriter(clientStream, Encoding.UTF8))
                {
                    String requestLine = fromClient.ReadLine();
                    Console.WriteLine("Request line: " + requestLine.Length + " " + requestLine);
                    if (String.IsNullOrWhiteSpace(requestLine))
                    {
                        SendError(toClient, 400, "Bad Request");
                    }
                    else
                    {
                        string[] parts = requestLine.Split(new[] { ' ' });
                        if (parts.Count() != 3)
                        {
                            SendError(toClient, 400, "Bad Request");
                        }
                        else
                        {
                            String method = parts[0];
                            String uri = parts[1];
                            String protocol = parts[2];
                            if (!IsProtocolLegal(protocol))
                            {
                                SendError(toClient, 400, "Illegal protocol");
                            }
                            else
                            {
                                switch (method)
                                {
                                    case "GET": DoGet(uri, toClient);
                                        break;
                                    case "POST":
                                    case "HEAD": SendError(toClient, 501, "Not Implemented");
                                        break;
                                    default: SendError(toClient, 400, "Bad Request");
                                        break;
                                }
                            }
                        }
                    }
                    toClient.Flush();
                }
            }
            finally { tcpClient.Close(); }
        }

        private static void DoGet(string uri, StreamWriter toClient)
        {
            try
            {
                string fullFilename = RootCatalog + uri;
                using (FileStream fileStream = new FileStream(fullFilename, FileMode.Open, FileAccess.Read))
                {
                    FileInfo fileInfo = new FileInfo(fullFilename);
                    long sizeInBytes = fileInfo.Length;
                    toClient.Write("HTTP/1.0 200 OK\r\n");
                    toClient.Write("Content-Length: {0}\r\n", sizeInBytes);
                    toClient.Write("\r\n");
                    toClient.Flush();
                    fileStream.CopyTo(toClient.BaseStream);
                }
            }
            catch (FileNotFoundException) { SendError(toClient, 404, "Not Found"); }
        }

        private static bool IsProtocolLegal(String protocol)
        {
            return "HTTP/1.1".Equals(protocol) || "HTTP/1.0".Equals(protocol);
        }

        private static void SendError(TextWriter toClient, int statusCode, string reasonPhrase)
        {
            String responseLine = string.Format("HTTP/1.0 {0} {1}\r\n", statusCode, reasonPhrase);
            toClient.Write(responseLine);
            toClient.Write("Content-Type: text/html\r\n");
            toClient.Write("\r\n");
            toClient.Write("<html><head><title>{0} {1}</title></head><body>{0} {1}</body></html>",
                 statusCode, reasonPhrase);
        }

        static void Main(string[] args)
        {
            HttpServer server = new HttpServer(DefaultPort);
            server.RunServer();
        }
    }
}